__title__ = "citros"
__description__ = "citros"
__url__ = "https://citros.io"
__version__ = "0.2.14"
__author__ = "lulav"
__author_email__ = "contact@lulav.space"
__license__ = "LICENSE"
__copyright__ = "Copyright 2024 Lulav Space"
